from .models import ChessBert, ChessGPT2, ChessT5
from ._loss import segmented_loss
from ._decode import segmented_argmax
from ._masking import random_masking
from . import chars, uci, fen, multi

__all__ = [
  'ChessBert', 'ChessGPT2', 'ChessT5',
  'segmented_loss',
  'segmented_argmax', 'random_masking',
  'chars', 'uci', 'fen', 'multi',
]


from typing import Sequence, TypedDict
import numpy as np
from haskellian import dicts as D
from ..metrics import accuracy, seq_accuracy, correct_count

class Metrics(TypedDict):
  acc: float
  seq_acc: float
  correct_len: int
  from_sq_acc: float
  to_sq_acc: float

def metrics(*, uci_preds: Sequence[str], uci_labs: Sequence[str]) -> Metrics:
  clean_preds = [p[:len(l)] for p, l in zip(uci_preds, uci_labs)]
  acc = accuracy(labs=uci_labs, preds=clean_preds)
  seq_acc = seq_accuracy(labs=uci_labs, preds=clean_preds)

  corr_len = correct_count(labs=uci_labs, preds=clean_preds)

  from_sq_preds = [p[0:2] for p in clean_preds]
  from_sq_labs = [l[0:2] for l in uci_labs]
  from_sq_acc = accuracy(labs=from_sq_labs, preds=from_sq_preds)

  to_sq_preds = [p[2:4] for p in clean_preds]
  to_sq_labs = [l[2:4] for l in uci_labs]
  to_sq_acc = accuracy(labs=to_sq_labs, preds=to_sq_preds)

  return Metrics(acc=acc, seq_acc=seq_acc, correct_len=corr_len, from_sq_acc=from_sq_acc, to_sq_acc=to_sq_acc)

def batch_metrics(*, uci_preds: Sequence[Sequence[str]], uci_labs: Sequence[Sequence[str]]) -> Metrics:
  sample_metrics = [metrics(uci_labs=labs, uci_preds=preds) for labs, preds in zip(uci_labs, uci_preds)]
  return D.aggregate(np.mean, sample_metrics) # type: ignore



from typing import Sequence, TypedDict
import numpy as np
from haskellian import dicts as D
from ..metrics import accuracy, seq_accuracy, correct_count

class Metrics(TypedDict):
  acc: float
  seq_acc: float
  correct_len: int
  fen_accs: Sequence[float]
  """Accuracy within each FEN (i.e. proportion of correct squares)"""

def metrics(*, fen_preds: Sequence[str], fen_labs: Sequence[str]) -> Metrics:
  return Metrics(
    acc=accuracy(labs=fen_labs, preds=fen_preds),
    seq_acc=seq_accuracy(labs=fen_labs, preds=fen_preds),
    correct_len=correct_count(labs=fen_labs, preds=fen_preds),
    fen_accs=[accuracy(labs=l, preds=p) for l, p in zip(fen_labs, fen_preds)]
  )

def batch_metrics(*, fen_preds: Sequence[Sequence[str]], fen_labs: Sequence[Sequence[str]]) -> Metrics:
  sample_metrics = [metrics(fen_labs=labs, fen_preds=preds) for labs, preds in zip(fen_labs, fen_preds)]
  return D.aggregate(np.mean, sample_metrics) # type: ignore



from ._labels import labels, parse_fen, decode_fen, Label, symbol_ids, piece_symbols
from .samples import parse_sample, Sample
from ._metrics import metrics, batch_metrics, Metrics

__all__ = [
  'labels', 'parse_fen', 'decode_fen', 'Label',
  'parse_sample', 'Sample', 'symbol_ids', 'piece_symbols',
  'metrics', 'batch_metrics', 'Metrics',
]


from typing import NamedTuple, Sequence
import chess_utils as cu
from ..vocab import Vocabulary
from ._labels import labels, Label

class Sample(NamedTuple):
  input_ids: Sequence[int]
  labs: Sequence[Label]

def parse_sample(
  inputs: str, ucis: str | None = None, *,
  vocab: Vocabulary, ignore_idx: int = -100
) -> Sample:
	sans = inputs.strip('\n').replace('+', '').replace('#', '').split(' ')
	input_ids = [vocab[san] for san in sans]
	parsed_ucis = cu.sans2ucis(sans) if ucis is None else ucis.strip('\n').split(' ')
	labs = labels(parsed_ucis, ignore_idx=ignore_idx)
	return Sample(input_ids, labs)


from typing import Sequence
from jaxtyping import Int, Float
import torch

def segmented_argmax(
  logits: Float[torch.Tensor, 'batch seq_len _'],
  *, ends: Sequence[int]
) -> Int[torch.Tensor, 'batch seq_len len(ends)']:
  split_logits = [logits[:, :, i:j] for i, j in zip([0, *ends[:-1]], ends)]
  argmaxs = [torch.argmax(split_logit, dim=-1) for split_logit in split_logits]
  return torch.stack(argmaxs, dim=-1)



from typing import Sequence, Protocol
from jaxtyping import Int, Float
import torch
from torch import Tensor, nn

class AggregateFn(Protocol):
  def __call__(self, x: Float[Tensor, 'segment batch']) -> Float[Tensor, 'batch']: ...

class LossFn(Protocol):
  def __call__(
    self, logits: Float[Tensor, 'batch seq_len _'],
    labels: Int[Tensor, 'batch seq_len']
  ) -> Float[Tensor, '']:
    ...


def segmented_loss(
  logits: Float[Tensor, 'batch seq_len _'],
  labels: Int[Tensor, 'batch seq_len len(ends)'],
  *,
  loss: LossFn = nn.CrossEntropyLoss(ignore_index=-100),
  ends: Sequence[int],
) -> list[Float[Tensor, '']]:
  split_logits = [logits[:, :, i:j].transpose(1, 2) for i, j in zip([0, *ends[:-1]], ends)]
  split_labels = [labels[:, :, i] for i in range(len(ends))]
  return [loss(logits, labs) for logits, labs in zip(split_logits, split_labels)]



from jaxtyping import Int
import torch
from torch import Tensor

def random_masking(
  batch: Int[Tensor, 'batch maxlen'],
  attn_mask: Int[Tensor, 'batch maxlen'] | None = None,
  *, mask_id: int, mask_prob: float
):
  """Mask `batch` with `mask_id` with probability `mask_prob`.
  - If provided, positions with `attn_mask == 0` are excluded from masking.
  """
  rand = torch.rand(batch.shape)
  indices = (rand < mask_prob) & (attn_mask is None or attn_mask == 1)
  output = batch.clone()
  output[indices] = mask_id
  return output



from typing import TypeAlias, Iterable, Sequence, Literal

promotion_ids = dict(n=0, b=1, r=2, q=3)
def encode_promotion(piece: Literal['n', 'b', 'r', 'q']) -> int:
  return promotion_ids[piece]

def encode_uci(e2e4q: str, *, ignore_idx: int = -100) -> tuple[int, int, int, int, int]:
  from_file = ord(e2e4q[0]) - ord('a')
  from_rank = int(e2e4q[1]) - 1
  to_file = ord(e2e4q[2]) - ord('a')
  to_rank = int(e2e4q[3]) - 1
  promotion = encode_promotion(e2e4q[4]) if len(e2e4q) == 5 else ignore_idx # type: ignore
  return from_file, from_rank, to_file, to_rank, promotion

Label: TypeAlias = tuple[int, int, int, int, int]

def labels(ucis: Iterable[str], *, ignore_idx: int = -100) -> Sequence[Label]:
  """`ignore_idx`: used for promotion when none"""
  return [encode_uci(uci, ignore_idx=ignore_idx) for uci in ucis]




from typing import Sequence
from jaxtyping import Float, Int
from haskellian import iter as I
import torch
from torch import Tensor, nn
from .._decode import segmented_argmax
from .._loss import segmented_loss
from ... import fen

ENDS = [13*(i+1) for i in range(64)]

def loss(
  logits: Float[Tensor, 'batch seq_len 64*13'],
  labels: Int[Tensor, 'batch seq_len 64'], *,
  ce_loss: nn.CrossEntropyLoss = nn.CrossEntropyLoss(ignore_index=-100)
) -> Float[Tensor, '']:
  """Cross-entropy loss summed across the board squares"""
  losses = segmented_loss(logits, labels, loss=ce_loss, ends=ENDS)
  return sum(l for l in losses if not torch.isnan(l)) # type: ignore

def argmax_fens(logits: Float[Tensor, 'B L 64*13']) -> list[list[str]]:
  argmaxs = segmented_argmax(logits, ends=ENDS)
  return I.ndmap(fen.decode_fen, argmaxs) # type: ignore

def metrics(sample_logits: Float[Tensor, 'L 64*13'], true_fens: Sequence[str]) -> fen.Metrics:
  pred_fens = argmax_fens(sample_logits)
  return fen.metrics(fen_labs=true_fens, fen_preds=pred_fens[0])


from typing import Sequence
from jaxtyping import Float, Int
from torch import Tensor, nn
import torch
from .._loss import segmented_loss
from ... import uci
from .decode import argmax_ucis

def loss(
  logits: Float[Tensor, 'batch seq_len 36'],
  labels: Int[Tensor, 'batch seq_len 5'], *,
  ce_loss: nn.CrossEntropyLoss = nn.CrossEntropyLoss(ignore_index=-100)
) -> Float[Tensor, '']:
  losses = segmented_loss(logits, labels, loss=ce_loss, ends=[8, 16, 24, 32, 36])
  return sum(l for l in losses if not torch.isnan(l)) # type: ignore
  # nan may happen if all promotion labels are -100 (which is quite likely)

def metrics(sample_logits: Float[Tensor, 'L 5'], true_ucis: Sequence[str]) -> uci.Metrics:
  pred_ucis = argmax_ucis(sample_logits)
  return uci.metrics(uci_labs=true_ucis, uci_preds=pred_ucis[0])


from . import pytorch, uci, fen, vocab, multi
from .metrics import accuracy, seq_accuracy, correct_count

__all__ = [
  'pytorch', 'fen', 'uci', 'vocab', 'multi',
  'accuracy', 'seq_accuracy', 'correct_count',
]


from typing import NamedTuple, Sequence
from haskellian import iter as I
from jaxtyping import Int
import torch
from ...uci import Sample as WordSample

class Sample(NamedTuple):
  input_ids: Int[torch.Tensor, 'L']
  labs: Int[torch.Tensor, 'L 5']
  uci_labs: Sequence[str]

class Batch(NamedTuple):
  input_ids: Int[torch.Tensor, 'B L']
  labs: Int[torch.Tensor, 'B L 5']
  uci_labs: Sequence[Sequence[str]]

def sample_from(sample: WordSample, uci_labs: Sequence[str]) -> Sample:
  return Sample(torch.tensor(sample.input_ids), torch.tensor(sample.labs), uci_labs)

def collate_fn(batch: Sequence[Sample], pad_token_id: int, ignore_idx: int = -100) -> Batch:

  input_ids, labs, uci_labs = I.unzip(batch)
  batch_size = len(input_ids)
  maxlen = max(len(x) for x in input_ids)
  padded_input_ids = torch.full((batch_size, maxlen), fill_value=pad_token_id)
  padded_labs = torch.full((batch_size, maxlen, 5), fill_value=ignore_idx)

  for i in range(batch_size):
    l = len(input_ids[i])
    padded_input_ids[i, :l] = input_ids[i]
    padded_labs[i, :l] = labs[i]

  return Batch(padded_input_ids, padded_labs, uci_labs)



from typing import Sequence, Mapping
from jaxtyping import Float
from torch import Tensor
from ..uci import metrics as uci_metrics
from ..fen import metrics as fen_metrics

def metrics(
  uci: Float[Tensor, 'L 5'],
  fen: Float[Tensor, 'L 64*13'],
  *,
  true_ucis: Sequence[str],
  true_fens: Sequence[str],
) -> Mapping[str, float | Sequence[float]]:
  return {
    f'uci-{k}': v for k, v in uci_metrics(uci, true_ucis).items()
  } | {
    f'fen-{k}': v for k, v in fen_metrics(fen, true_fens).items()
  } # type: ignore


from typing import Mapping
from jaxtyping import Int
import torch
from torch import nn, Tensor
from transformers import BertConfig, BertModel

class ChessBert(nn.Module):
  def __init__(
    self, *, vocab_size: int, max_len: int = 512,
    hidden_size: int = 768, attention_heads: int = 12,
    pad_token_id: int, heads: Mapping[str, int] = { 'uci': 36 }
  ):
    super().__init__()
    # BERT configuration
    self.config = BertConfig(
      vocab_size=vocab_size,
      max_position_embeddings=max_len,
      hidden_size=hidden_size,
      num_attention_heads=attention_heads,
      pad_token_id=pad_token_id
    )
    self.bert = BertModel(self.config)
    self.heads = nn.ModuleDict({
      head: nn.Linear(hidden_size, num_classes)
      for head, num_classes in heads.items()
    })

  def forward(
    self, input_ids: Int[Tensor, 'batch maxlen'],
    *, attention_mask: Int[Tensor, 'batch maxlen'] | None = None
  ):
    # Get BERT outputs
    outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask)
    sequence_output = outputs.last_hidden_state  # Shape: (batch_size, seq_len, hidden_size)
    return { head: self.heads[head](sequence_output) for head in self.heads }
  

  def insert_token(self, idx: int):
    """Insert new token in the model's vocabulary, at the given `idx`"""
    config = self.config
    new_vocab_size = config.vocab_size + 1
    new_word_embed = nn.Embedding(new_vocab_size, config.hidden_size, padding_idx=config.pad_token_id)
    old_embed = self.bert.embeddings.word_embeddings.weight
    with torch.no_grad():
      new_word_embed.weight[:idx] = old_embed[:idx]
      new_word_embed.weight[idx] = torch.mean(old_embed, dim=0)
      new_word_embed.weight[idx+1:] = old_embed[idx:]
    
    config.vocab_size = new_vocab_size
    self.config = config
    self.bert.embeddings.word_embeddings = new_word_embed

  def append_token(self):
    """Append new token in the model's vocabulary"""
    self.insert_token(self.config.vocab_size-1)


from typing import TypedDict, Sequence
import chess_utils as cu
from ..vocab import Vocabulary
from .. import fen, uci

class Sample(TypedDict):
  input_ids: Sequence[int]
  ucis: Sequence[uci.Label]
  fens: Sequence[fen.Label]

def parse_sample(
  inputs: str, ucis: str | None = None, fens: str | None = None, *,
  vocab: Vocabulary, ignore_idx: int = -100
) -> Sample:
  sans = inputs.strip().replace('+', '').replace('#', '').split()
  input_ids = [vocab[san] for san in sans]
  parsed_ucis = cu.sans2ucis(sans) if ucis is None else ucis.strip().split()
  uci_labs = uci.labels(parsed_ucis, ignore_idx=ignore_idx)
  parsed_fens = cu.sans2fens(sans) if fens is None else fens.strip().split()
  fen_labs = fen.labels(parsed_fens)
  return Sample(input_ids=input_ids, ucis=uci_labs, fens=fen_labs)